/* include/pqxx/config.h.in.  Generated from configure.ac by autoheader.  */
/* define if the compiler supports basic C++17 syntax */
/* #undef HAVE_CXX17 */
/* Define to 1 if you have the <dlfcn.h> header file. */
/* #undef HAVE_DLFCN_H */
/* Define to 1 if you have the <inttypes.h> header file. */
/* #undef HAVE_INTTYPES_H */
/* Define to 1 if you have the `pq' library (-lpq). */
/* #undef HAVE_LIBPQ */
/* Define to 1 if you have the <memory.h> header file. */
/* #undef HAVE_MEMORY_H */
/* Define to 1 if you have the <stdint.h> header file. */
/* #undef HAVE_STDINT_H */
/* Define to 1 if you have the <stdlib.h> header file. */
/* #undef HAVE_STDLIB_H */
/* Define to 1 if you have the <strings.h> header file. */
/* #undef HAVE_STRINGS_H */
/* Define to 1 if you have the <string.h> header file. */
/* #undef HAVE_STRING_H */
/* Define to 1 if you have the <sys/stat.h> header file. */
/* #undef HAVE_SYS_STAT_H */
/* Define to 1 if you have the <sys/types.h> header file. */
/* #undef HAVE_SYS_TYPES_H */
/* Define to 1 if you have the <unistd.h> header file. */
/* #undef HAVE_UNISTD_H */
/* Define to the sub-directory where libtool stores uninstalled libraries. */
/* #undef LT_OBJDIR */
/* Name of package */
/* #undef PACKAGE */
/* Define to the address where bug reports for this package should be sent. */
/* #undef PACKAGE_BUGREPORT */
/* Define to the full name of this package. */
/* #undef PACKAGE_NAME */
/* Define to the full name and version of this package. */
/* #undef PACKAGE_STRING */
/* Define to the one symbol short name of this package. */
/* #undef PACKAGE_TARNAME */
/* Define to the home page for this package. */
/* #undef PACKAGE_URL */
/* Define to the version of this package. */
/* #undef PACKAGE_VERSION */
/* Define if <charconv> supports floating-point conversion. */
/* #undef PQXX_HAVE_CHARCONV_FLOAT */
/* Define if <charconv> supports integer conversion. */
#define PQXX_HAVE_CHARCONV_INT
/* Define if compiler supports Concepts and <ranges> header. */
/* #undef PQXX_HAVE_CONCEPTS */
/* Define if compiler supports __cxa_demangle */
/* #undef PQXX_HAVE_CXA_DEMANGLE */
/* Define if g++ supports pure attribute */
/* #undef PQXX_HAVE_GCC_PURE */
/* Define if g++ supports visibility attribute. */
/* #undef PQXX_HAVE_GCC_VISIBILITY */
/* Define if likely & unlikely work. */
#define PQXX_HAVE_LIKELY
/* Define if poll() is available. */
/* #undef PQXX_HAVE_POLL */
/* Define if libpq has PQencryptPasswordConn (since pg 10). */
/* #undef PQXX_HAVE_PQENCRYPTPASSWORDCONN */
/* Define if std::this_thread::sleep_for works. */
#define PQXX_HAVE_SLEEP_FOR
/* Define if compiler has std::span. */
/* #undef PQXX_HAVE_SPAN */
/* Define if thread_local is fully supported. */
#define PQXX_HAVE_THREAD_LOCAL
/* Define to 1 if you have the ANSI C header files. */
/* #undef STDC_HEADERS */
/* Version number of package */
/* #undef VERSION */
