



if not exist %LIBRARY_INC%\pqxx\pqxx (exit 1)
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_LIB%\pqxx.lib (exit 1)
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_BIN%\pqxx.dll (exit 1)
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
